/*
 * $Id: invoke-main.c,v 1.2.2.7 1997/12/02 16:04:49 randall Exp $
 */

/*
 * $Log: invoke-main.c,v $
 * Revision 1.2.2.7  1997/12/02 16:04:49  randall
 * Bug fix.
 *
 * Revision 1.2.2.6  1997/11/10 22:27:57  randall
 * Updated some interfaces, moved some more socket initialization
 * into message-udp.c
 *
 * Revision 1.2.2.5  1997/11/01 07:31:44  randall
 * Got critical path/work instrumentation to work.  Fixed other small bugs.
 *
 * Revision 1.2.2.4  1997/10/23 03:00:22  randall
 * Small fix to handle James' message-passing library.
 *
 * Revision 1.2.2.3  1997/10/21 23:02:42  randall
 * Separated out UDP implementation.  Reorganized caching algorithm.
 *
 * Revision 1.2.2.2  1997/10/14 20:26:04  randall
 * First working version of distributed cilk on DEC machines.
 *
 * Revision 1.2.2.1  1997/06/13 22:33:21  randall
 * Added distributed version to the repository under
 * branch "distributed"
 *
 * Revision 1.2  1997/03/01 17:42:54  athena
 * Massive cleanup and merging --- since I was still using the
 * old repository :-(
 *
 */

/*
 * Copyright (c) 1996 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to use, copy, modify, and distribute the Software without
 * restriction, provided the Software, including any modified copies made
 * under this license, is not distributed for a fee, subject to
 * the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE MASSACHUSETTS INSTITUTE OF TECHNOLOGY BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
 * CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * Except as contained in this notice, the name of the Massachusetts
 * Institute of Technology shall not be used in advertising or otherwise
 * to promote the sale, use or other dealings in this Software without
 * prior written authorization from the Massachusetts Institute of
 * Technology.
 *  
 */

#include <cilk.h>
#include <cilk-lib.h>
#include <cilk-internal.h>
#include <cilk-cilk2c.h>
#include <stdio.h>

DECLARE_SHARED_VARIABLE(extern int, done, );

/************************************************************
 *  Setting done flag
 ************************************************************/

typedef struct
{
  int retval;
  WHEN_CILK_TIMING(Cilk_time total_work;)
  WHEN_CILK_TIMING(Cilk_time critical_path;)
} DoneRequest;

void set_done(DoneRequest *req)
{
  WHEN_CILK_TIMING(USE_MACHINE(total_work) = req->total_work);
  WHEN_CILK_TIMING(USE_MACHINE(critical_path) = req->critical_path);
  USE_MACHINE(retval) = req->retval;
  Cilk_membar_StoreStore();
  USE_MACHINE(done) = 1;
}

/************************************************************
 *  Initial thread
 ************************************************************/

/* This is a hand-compiled procedure that calls cilk_main */
struct invoke_main_frame {
     StackFrame header;
     int argc;
     char **argv;
     int retval;
};

/* you are not supposed to understand this */
extern int cilk_main(int argc, char *argv[]);
static void invoke_main_slow(struct invoke_main_frame *_frame)
{
     int argc;
     char **argv;
     int tmp;
     int i;

     _START_THREAD_SLOW();
     switch (_frame->header.entry) {
	 case 1:
	      goto _sync1;
	 case 2:
	      goto _sync2;
     }

     /* set up shared-memory segment correctly */
     {
       void *copy;
       size_t size;
       void Cilk_unmap_region(void *addr, size_t size);
       
       CILK_ASSERT(Self == 0);
       
       size = USE_PRIVATE(shared_region_end) - USE_PRIVATE(shared_region_begin);
       if (size > 0)
	 {
	   copy = malloc(size);
	   CILK_ASSERT(copy != NULL);
	   memcpy(copy, USE_PRIVATE(shared_region_begin), size);
	   Cilk_unmap_region(USE_PRIVATE(shared_region_begin), size);
	   memcpy(USE_PRIVATE(shared_region_begin), copy, size);
	   free(copy);
	 }
     }

     Cilk_user_work = 0;
     Cilk_user_critical_path = 0;
     WHEN_CILK_TIMING(_frame->header.cp = (Cilk_time) 0);
     WHEN_CILK_TIMING(_frame->header.work = (Cilk_time) 0);
     WHEN_CILK_TIMING(_frame->header.mycp = (Cilk_time) 0);
     WHEN_CILK_TIMING(USE_PRIVATE(last_cp_time) = Cilk_get_time());

     argc = _frame->argc;
     argv = _frame->argv;

     _frame->header.index = (char *) &_frame->retval -
	  (char *) _frame;
     _frame->header.entry = 1;
     _BEFORE_SPAWN_SLOW();
     _PUSH_FRAME(_frame);
     tmp = cilk_main(argc, argv);
     _XPOP_FRAME_RESULT(_frame, /* return nothing */ , tmp);
     _AFTER_SPAWN_SLOW();
     _frame->retval = tmp;

     if (0) {
	_sync1:;
     }
     _AT_THREAD_BOUNDARY_SLOW();

     _BEFORE_SYNC_SLOW();
     _frame->header.entry = 2;
     if (Cilk_sync()) {
	  return;
	_sync2:;
     }
     _AFTER_SYNC_SLOW();
     _AT_THREAD_BOUNDARY_SLOW();

     fflush(NULL);

     for(i=0; i<USE_MACHINE(Machines); i++)
       {
	 DoneRequest *req;
	 req = Cilk_alloc_msg_buf(sizeof(DoneRequest), i);
	 req->retval = _frame->retval;
	 WHEN_CILK_TIMING(req->total_work = _frame->header.work);
	 WHEN_CILK_TIMING(req->critical_path = _frame->header.mycp);
	 /* note: we special-case the local set_done call to make
	  * sure the done variable gets set immediately on the local machine.
	  */
	 if (i == USE_MACHINE(Machine))
	   {
	     set_done(req);
	     Cilk_free_msg_buf(req, sizeof(DoneRequest), i);
	   }
	 else
	   Cilk_rpc(i, set_done, req, sizeof(*req));
       }
     
     Cilk_internal_free_fast(_frame, sizeof(struct invoke_main_frame));
     return;
}

static ProcInfo invoke_main_sig[] =
{
     {0, 0, invoke_main_slow, 0, 0},
     {sizeof(int), offsetof(struct invoke_main_frame, retval), 0, 0, 0},
     {0, 0, 0, 0, 0}
};

Closure *Cilk_create_initial_thread(int argc, char *argv[])
{
     Closure *t;
     struct invoke_main_frame *f;

     /* create a frame for invoke_cilk_main */
     t = Cilk_Closure_create();
     t->parent = (ClosureID){NULL, -1};
     t->join_counter = 0;
     t->child_num = -1;
     t->curr_child = 1;
     t->status = CLOSURE_READY;

     f = Cilk_internal_malloc_fast(sizeof(struct invoke_main_frame));
     f->header.entry = 0;
     f->header.sig = invoke_main_sig;
     WHEN_CILK_DEBUG(f->header.magic = CILK_STACKFRAME_MAGIC);

     f->argc = argc;
     f->argv = argv;

     t->frame = &f->header;
     return t;
}
